@@ -5,7 +5,7 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
-;; Version: 2.1
+;; Version: 2.5
 ;; Package-Requires: ((emacs "28.1") (compat "30"))
 ;; URL: https://github.com/minad/vertico
 ;; Keywords: convenience, files, matching, completion
@@ -56,7 +56,7 @@
 (defcustom vertico-group-format
   (concat #("    " 0 4 (face vertico-group-separator))
           #(" %s " 0 4 (face vertico-group-title))
-          #(" " 0 1 (face vertico-group-separator display (space :align-to right))))
+          #(" " 0 1 (face vertico-group-separator display (space :align-to (- right 1)))))
   "Format string used for the group title."
   :type '(choice (const :tag "No group titles" nil) string))
 
@@ -144,6 +144,13 @@ The value should lie between 0 and vertico-count/2."
   "TAB" #'vertico-insert
   "<touchscreen-begin>" #'ignore)
 
+(defvar vertico--locals
+  '((scroll-margin . 0)
+    (completion-auto-help . nil)
+    (completion-show-inline-help . nil)
+    (pixel-scroll-precision-mode . nil))
+  "Vertico minibuffer local variables.")
+
 (defvar-local vertico--hilit #'identity
   "Lazy candidate highlighting function.")
 
@@ -463,14 +470,44 @@ The value should lie between 0 and vertico-count/2."
         (put-text-property beg next 'face (remq face (ensure-list val)) obj))
       (setq beg next))))
 
+(defun vertico--debug (&rest _)
+  "Debugger used by `vertico--protect'."
+  (let ((inhibit-message t))
+    (require 'backtrace)
+    (declare-function backtrace-to-string "backtrace")
+    (message "Vertico detected an error:\n%s" (backtrace-to-string)))
+  (let (message-log-max)
+    (message "%s %s"
+             (propertize "Vertico detected an error:" 'face 'error)
+             (substitute-command-keys "Press \\[view-echo-area-messages] to see the stack trace")))
+  nil)
+
+(defun vertico--protect (fun)
+  "Protect FUN such that errors are caught.
+If an error occurs, the FUN is retried with `debug-on-error' enabled and
+the stack trace is shown in the *Messages* buffer."
+  (static-if (fboundp 'handler-bind) ;; Available on Emacs 30
+      (ignore-errors
+        (handler-bind ((error #'vertico--debug))
+          (funcall fun)))
+    (when (or debug-on-error (condition-case nil
+                                 (progn (funcall fun) nil)
+                               (error t)))
+      (let ((debug-on-error t)
+            (debugger #'vertico--debug))
+        (condition-case nil
+            (funcall fun)
+          ((debug error) nil))))))
+
 (defun vertico--exhibit ()
   "Exhibit completion UI."
-  (let ((buffer-undo-list t)) ;; Overlays affect point position and undo list!
-    (vertico--update 'interruptible)
-    (vertico--prompt-selection)
-    (vertico--display-count)
-    (vertico--display-candidates (vertico--arrange-candidates))
-    (vertico--resize)))
+  (vertico--protect
+   (lambda ()
+     (let ((buffer-undo-list t)) ;; Overlays affect point position and undo list!
+       (vertico--update 'interruptible)
+       (vertico--prompt-selection)
+       (vertico--display-count)
+       (vertico--display-candidates (vertico--arrange-candidates))))))
 
 (defun vertico--goto (index)
   "Go to candidate with INDEX."
@@ -551,20 +588,24 @@ The value should lie between 0 and vertico-count/2."
 (cl-defgeneric vertico--display-candidates (lines)
   "Update candidates overlay `vertico--candidates-ov' with LINES."
   (move-overlay vertico--candidates-ov (point-max) (point-max))
-  (overlay-put vertico--candidates-ov 'after-string
-               (apply #'concat #(" " 0 1 (cursor t)) (and lines "\n") lines)))
+  (overlay-put vertico--candidates-ov 'before-string
+               (apply #'concat #(" " 0 1 (cursor t)) (and lines "\n") lines))
+  (vertico--resize-window (length lines)))
 
-(cl-defgeneric vertico--resize ()
-  "Resize active minibuffer window."
+(cl-defgeneric vertico--resize-window (height)
+  "Resize active minibuffer window to HEIGHT."
   (setq-local truncate-lines (< (point) (* 0.8 (vertico--window-width)))
-              resize-mini-windows vertico-resize
+              resize-mini-windows 'grow-only
               max-mini-window-height 1.0)
   (unless truncate-lines (set-window-hscroll nil 0))
-  (unless (or vertico-resize (frame-root-window-p (active-minibuffer-window)))
-    (let ((delta (- (max (cdr (window-text-pixel-size))
-                         (* (default-line-height) (1+ vertico-count)))
+  (unless (frame-root-window-p (active-minibuffer-window))
+    (unless vertico-resize (setq height (max height vertico-count)))
+    (let ((dp (- (max (cdr (window-text-pixel-size))
+                      (* (default-line-height) (1+ height)))
                  (window-pixel-height))))
-      (when (/= 0 delta) (window-resize nil delta nil nil 'pixelwise)))))
+      (when (or (and (> dp 0) (/= height 0))
+                (and (< dp 0) (eq vertico-resize t)))
+        (window-resize nil dp nil nil 'pixelwise)))))
 
 (cl-defgeneric vertico--prepare ()
   "Ensure that the state is prepared before running the next command."
@@ -575,12 +616,9 @@ The value should lie between 0 and vertico-count/2."
 
 (cl-defgeneric vertico--setup ()
   "Setup completion UI."
-  (when (boundp 'pixel-scroll-precision-mode)
-    (setq-local pixel-scroll-precision-mode nil))
-  (setq-local scroll-margin 0
-              vertico--input t
-              completion-auto-help nil
-              completion-show-inline-help nil
+  (dolist (var vertico--locals)
+    (set (make-local-variable (car var)) (cdr var)))
+  (setq-local vertico--input t
               vertico--candidates-ov (make-overlay (point-max) (point-max) nil t t)
               vertico--count-ov (make-overlay (point-min) (point-min) nil t t))
   (overlay-put vertico--count-ov 'priority 1) ;; For `minibuffer-depth-indicate-mode'
@@ -590,7 +628,8 @@ The value should lie between 0 and vertico-count/2."
 
 (cl-defgeneric vertico--advice (&rest app)
   "Advice for completion function, apply APP."
-  (minibuffer-with-setup-hook #'vertico--setup (apply app)))
+  (dlet ((completion-eager-display nil)) ;; Available on Emacs 31
+    (minibuffer-with-setup-hook #'vertico--setup (apply app))))
 
 (defun vertico-first ()
   "Go to first candidate, or to the prompt when the first candidate is selected."
